function [D, w, acc_train, acc_test] = last(X, y, Omega_ini, s, numIterMax, params, X_test, y_test)
    % The function runs LAST algorithm on the set of training samples, and
    % returns the trained parameters (D, w).
    % Input:
      % X: Datapoints (as columns)
      % y: labels
      % Omega_ini: initial point (row concatenation of U_0 and v_0')
      % s: prior knowledge on the signs of w
      % params (all parameters are optional)
            % .alpha: alpha (default: 1) 
            % .beta: beta (default: 100) 
            % .nu: regularization parameter nu (default: 1)
            % .visualize: (default: 0)
            % .iters_sgd: number of iterations of SGD (default: 1000)
            % .mini_batch_size: batch_size (default: -1 (i.e., batch gradient descent))
            % .u0: number of steps where the step size is constant (default: params.iters_sgd/10)
            % .lambdas: lambdas to test in cross-validation region (default: 10.^(-(1:3)))
            % .cross_validation_region: region to test the different lambdas (default: params.u0/2)
            % .delta: termination criterion (see paper for more details) (default: 0 (i.e., will do numIterMax iterations)) 
      % optional arguments: X_test, y_test to track the test accuracy
            
    % Output: 
      % D: dictionary
      % w: normal vector of linear separator
      % acc_train: accuracy of training
      % acc_test: accuracy of testing
      
    %  Alhussein Fawzi.
    %  Signal Processing Laboratory (LTS4)
    %  EPFL, Lausanne, Switzerland.
    %  alhussein.fawzi@epfl.ch
    %
    %  February 2014

    % Reference:
    % A. Fawzi, M. Davies, P. Frossard, "Dictionary learning for fast 
    % classification based on soft-thresholding", International Journal on
    % Computer Vision (IJCV), 2014.

    Omega = Omega_ini;

    n = size(X, 1);
    
    if (nargin < 6)
        fprintf('Incorrect number of parameters\n');
    elseif (nargin == 6)
        ok_test = 0;
        acc_test = -1;
    else
        ok_test = 1;
    end
    
    % Check for parameters
    if (~isfield(params,'alpha'))
        params.alpha = 1;
    end
    
    if (~isfield(params,'beta'))
        params.beta = 100;
    end
    
    if (~isfield(params,'nu'))
        params.nu = 1;
    end
    
    if (~isfield(params,'visualize'))
        params.visualize = 0;
    end
    
    if (~isfield(params,'iters_sgd'))
        params.iters_sgd = 1000;
    end
    
    if (~isfield(params,'mini_batch_size'))
        params.mini_batch_size = -1;
    end
    
    if (~isfield(params,'u0'))
        params.u0 = params.iters_sgd/10;
    end
    
    if (~isfield(params,'lambdas'))
        params.lambdas = 10.^(-(1:3));
    end

    if (~isfield(params,'cross_validation_region'))
        params.cross_validation_region = params.u0/2;
    end
    
    if (~isfield(params,'delta'))
        params.delta = 0;
    end
    
    for k = 1:numIterMax
        fprintf('Iteration %d\n', k);
        grad_H = gradient_h(X, y, s, Omega, params);
        Omega_old = Omega;

        % Stochastic gradient descent
        tic;
        Omega = sgd(X, y, s, Omega, grad_H, params);
        time_sgd = toc;
        fprintf('Took %f seconds to do the sgd\n', time_sgd);

        % Compute objective of linearized function
        obj_linearized = compute_obj_function_linearized(X, y, s, Omega, Omega_old, grad_H, params);

        % Compute objective function of true function
        obj = compute_obj_function(X, y, s, Omega, params);
        fprintf('Obj. fun.: %f\n', obj);
        
        U = Omega(1:n, :);
        v = Omega(n+1, :)';
        w = s(:) .* v(:);
        D = U * diag(1./v);
        
        % obj_originalFormulation = compute_obj_function_originalFormulation(X, y, D, w, params.nu, params.alpha, params.beta);
        % fprintf('Obj. fun. original formulation: %f\n', obj_originalFormulation);
              
        criterion_matrix = min(abs(Omega - Omega_old), abs(Omega - Omega_old./Omega_old));
        
        % Compute train and test feature vectors
        % Train
        fea_train = max(0, D'*X - params.alpha);
        est_labels = w'*fea_train; est_labels(est_labels > 0) = 1; est_labels(est_labels <= 0) = -1;
        acc_train = 1 - compute_error_rate(est_labels(:), y(:));
        fprintf('Accuracy training = %f\n', acc_train);
        
        if (ok_test)
            % Test
            fea_test = max(0, D'*X_test - params.alpha);
            est_labels = w'*fea_test; 
            if(params.visualize == 1)
                figure; stem(est_labels); hold on; stem(y_test, 'r'); title('Testing accuracy');
            end
            
            est_labels(est_labels > 0) = 1; est_labels(est_labels <= 0) = -1;
            acc_test = 1 - compute_error_rate(est_labels(:), y_test(:));
            fprintf('Accuracy testing = %f\n', acc_test);
        end
        
        % Termination criterion of DCA
        if (all(all(criterion_matrix < params.delta)))
            fprintf('I reached a local optimum.\n');
            break;
        end
        
        if (obj < 1e-12)
            fprintf('Reached the global minimum.. Quitting\n');
            break;
        end
        
       %  Check for errors (debug)
        if (obj-obj_linearized > 1e-10)
           error('!!! Problem: the linearized objective function is smaller than the objective function itself... Breaking. !!! \n');
        end

    end
    
end