function Omega = sgd(X, y, s, Omega_ini, grad_H, params)
    
    % Input:
        % X: Datapoints (as columns)
        % y: labels
        % s: signs of w
        % Omega_ini: starting point
        % grad_H: gradient of h, around which we want to linearize
        % params (all required):
            % .alpha: alpha; .beta: beta, .nu: nu,
            % .visualize: visualize
            % .iters_sgd: number of iterations of SGD
            % .mini_batch_size: batch_size
            % .u0: number of steps until where the step size is constant
            % .lambdas: lambdas to test in cross-validation region
            % .cross_validation_region: region to test the different
            % lambdas
            
    % Output: the new Omega

    %  Alhussein Fawzi.
    %  Signal Processing Laboratory (LTS4)
    %  EPFL, Lausanne, Switzerland.
    %  alhussein.fawzi@epfl.ch
    %
    %  February 2014
    
    n = size(X, 1);
    m = size(X, 2);
    lambdas = params.lambdas;
    Omega = cell(1,length(lambdas));
    Omega_grad = cell(1, length(lambdas));
    
    if (params.mini_batch_size == -1)
        % Batch gradient descent
        mini_batch_size = m;
    else
        mini_batch_size = params.mini_batch_size; 
    end
    
    for q = 1:length(lambdas)
        Omega{q} = Omega_ini;
        objs{q} = []; % remove that 
    end
    
    X_cur = X; y_cur = y;
    rand_idxs = [];
    for k = 1:ceil((params.iters_sgd * mini_batch_size)/m)
        rand_idxs = [rand_idxs, randperm(m)];
    end
    
    if (params.visualize == 1) 
        cur_fig = figure;
    end
    
    for u = 1:params.iters_sgd
        if (mini_batch_size < m)
            % SGD (if batch GD, no need to reshuffle)
            cur_batch = rand_idxs(((u-1)*mini_batch_size+1):(u*mini_batch_size));
            X_cur = X(:, cur_batch); y_cur = y(cur_batch);
        end
        
        if (u == params.cross_validation_region+1)
           best_idx_lambda = 0; score_best_lambda = Inf;
           % Stop cross validation and maintain only the best lambda
           for q = 1:length(lambdas)
               cur_obj = compute_g(X, y, s, Omega{q}, params) - trace(Omega{q}' * grad_H);
               if (cur_obj < score_best_lambda)
                   score_best_lambda = cur_obj;
                   best_idx_lambda = q;
               end
           end
           fprintf('Chosen lambda with index = %f\n', best_idx_lambda);
           % Remove other lambdas
           lambdas(1) = lambdas(best_idx_lambda); lambdas(2:end) = [];
           Omega{1} = Omega{best_idx_lambda}; Omega(2:end) = [];
           objs{1} = objs{best_idx_lambda}; objs(2:end) = []; % to remove
        end
        for q = 1:length(lambdas)
            cur_grad = gradient_g(X_cur, y_cur, s, Omega{q}, params);
            cur_grad(n+1, :) = cur_grad(n+1, :) + mini_batch_size / m * params.nu * Omega{q}(n+1, :);
            cur_grad = cur_grad - (mini_batch_size / m) * grad_H;
            Omega_grad{q} = cur_grad;
        end
        
        for q = 1:length(lambdas)
            Omega{q} = Omega{q} - min(lambdas(q), lambdas(q) * params.u0/u) * Omega_grad{q};
            % Project on v >= epsilon
            Omega{q}(n+1, :) = max(1e-15, Omega{q}(n+1, :));
        end
        
        if (params.visualize == 1 && u > params.cross_validation_region)
            if (mod(u, 100) == 0)
                fprintf('>> sgd iter = %d\n', u);
                for q = 1:length(lambdas)
                    objs{q} = [objs{q}, compute_g(X, y, s, Omega{q}, params) - trace(Omega{q}' * grad_H)];
                    figure(cur_fig); hold on; plot(objs{q}, 'r');
                end
            end
        end
    end
    
    Omega = Omega{1};
end